; Color Quantizing image editor
; reduce the number of unique colors in an image
; show histograms of complete image or selected area
; display and edit image palettes
; other image editing functions
; Name     : ColorQuantizer.pb
; Author   : BasicallyPure
; Compiler : PureBasic 5.22 LTS
; License  : free
; Platform : Windows, Linux, Mac?

EnableExplicit
#Version = "1.00" ; 5/24/2014

; STOP AND READ!!! -- you need this module!!! --
; forum topic: http://www.purebasic.fr/english/viewtopic.php?f=40&t=57939

IncludeFile "NeuQuant_Module.pbi"
UseModule NEUQUANT

#PopMenu = 0
#UndoRef = %0001
#RedoRef = %0010
#UndoWrk = %0100
#RedoWrk = %1000

;{ declarations
Declare INIT_GUI()
Declare EVENT_LOOP()
Declare UNDO()
Declare REDO()
Declare CANVAS_EVENT_HANDLER()
Declare COUNT_COLORS(image.i)
Declare MN(n)
Declare SHOW_PALETTE(Array palette(1))
Declare EDIT_PALETTE_COLOR()
Declare SCAN_FOR_PALETTE()
Declare TOG(option)
Declare UPDATE_STATUS()
Declare COMPUTE_HISTOGRAM(image.i)
Declare DISPLAY_HISTOGRAM(image.i)
Declare CHOOSE_HISTOGRAM()
Declare CANCEL_SELECTION()
Declare SET_UNDO_REDO_BUTTONS(mode)
Declare LOAD_IMAGE()
Declare SAVE_IMAGE()
Declare REVERT()
Declare NEW_IMAGE()
Declare COPY()
Declare PASTE()
Declare RESIZE()
Declare ADD_TEXT()
Declare UPDATE_CANVAS(image)
Declare PREPARE_UNDO(image)
Declare FLIP(image)
Declare MIRROR(image)
Declare ROTATE_90(image,dir)
Declare NEGATIVE(image)
Declare BLACK_WHITE(image)
Declare BW_DITHER(image)
Declare ADJUST_CONTRAST()
Declare ADJUST_BRIGHTNESS()
Declare GRAY(image)
Declare CROP(image)
Declare SET_AS_REFERENCE()
Declare CLOSE_PWIN()
Declare CONFIG_Q_LIMIT()
Declare Q_LIMIT_ADJUST()
Declare Q_LIMIT_ENTER_KEY()
Declare WINDOW_MOVED()
Declare EXECUTE_QUANTIZATION()
Declare POPULARITY_PALETTE()
Declare UNIFORM_PALATTE()
Declare ADAPTIVE_PALETTE()
Declare FIXED_PALETTE(palette)
Declare NEUQUANT_256()
Declare BIT_MASK(option)
Declare ASSEMBLE_TO_PALETTE(Array palette(1))
Declare SET_SELECTION_BOUNDARIES(canvas)
Declare SELECT_AREA(canvas) : ;}

;{ enumerations
Enumeration ; windows
   #WinMain : #WinHistogram : #WinPalette : #WinContrast : #WinBright : #WinNew
   #WinText
EndEnumeration

Enumeration ; images
   #ImgMaster  : #ImgArrow_R : #ImgRef : #imgUndoWrk : #ImgWork : #ImgHist
   #imgPalette : #ImgArrow_L : #ImgSel : #imgUndoRef : #ImgTemp
EndEnumeration

Enumeration ; gadgets
   #BB     : #Q_limit  : #Canvas : #Btn_Redo : #Chk_Dither  : #ScrollArea
   #status : #Chk_Hist : #cc     : #Btn_Undo : #Combo_Qmode : #Opt_ViewRef
   #ChkRed : #ChkGrn   : #ChkBlu : #ChkExEx  : #ImgGadPal   : #Btn_Apply
   #Opt_ViewWrk
EndEnumeration

Enumeration #PB_Event_FirstCustomValue
   #CanvasMouseUP : #CanvasMouseDown : #CanvasRightClick : #CanvasRedo
   #CanvasCopy    : #CanvasPaste     : #CanvasUndo
EndEnumeration

Enumeration ; memu items
   #MenuItem_Revert : #MenuItem_Flip  : #MenuItem_Mirror  : #MenuItem_Negative
   #MenuItem_Rot_R  : #MenuItem_Crop  : #MenuItem_Gray    : #MenuItem_BW_Dither
   #MenuItem_BW     : #MenuItem_Rot_L : #MenuItem_Set_Ref : #MenuItem_Load
   #MenuItem_Save   : #MenuItem_Copy  : #MenuItem_Paste   : #MenuItem_Contrast
   #MenuItem_Enter  : #MenuItem_New   : #MenuItem_AddText : #MenuItem_Brightness
   #MenuItem_LArrow : #MenuItem_RArrow: #MenuItem_UArrow  : #MenuItem_DArrow
EndEnumeration

Enumeration ; quantization styles
   #Q_popular      : #Q_uniform      : #Q_adaptive     : #BMSK_Flat_8
   #BMSK_flat_64   : #BMSK_Flat_512  : #BMSK_Flat_4096 : #BMSK_lemon_128
   #BMSK_berry_128 : #BMSK_ocean_128 : #Q_Fixed_16
   #Q_fixed_24     : #Q_fixed_32     : #Q_neuquqnt_256
EndEnumeration : ;}

;{ structures
Structure HistogramType
   R.i[256]
   G.i[256]
   B.i[256]
   maxR.i
   maxG.i
   maxB.i
EndStructure

Structure SelType
   left.i   ; left edge 'x' value of selected area
   right.i  ; right edge 'x' value of selected area
   top.i    ; top edge 'y' value of selected area
   bottom.i ; bottom edge 'y' value of selected area
   active.i ; set to #False when image is new or updated
   sFlag.i  ; if #True an area is selected
   limLeft.i   ; left boundary limit of selection area (window coordinates)
   limRight.i  ; right boundary limit of selection area (window coordinates)
   limTop.i    ; top boundary limit of selection area (window coordinates)
   limBottom.i ; bottom boundary limit of selection area (window coordinates)
EndStructure

Structure PopType ; used in POPULARITY_PALETTE()
   clr.l ; color
   pop.i ; popularity
EndStructure : ;}

;{ global variables
Global Hist.HistogramType
Global selection.SelType
Global Dim WK_Pal(0)
Global SourceFileName$, File$
Global ImgActive = #ImgRef
Global ControlKey_Down
Global UndoRedoStates
Global EndProgram = #False
;}

If INIT_GUI()
   EVENT_LOOP()
EndIf

End

Procedure INIT_GUI()
   Protected c, r, g, b, x, y, h = 620, w = 800
   Protected flags = #PB_Window_ScreenCentered | #PB_Window_SystemMenu |
             #PB_Window_MinimizeGadget | #PB_Window_MaximizeGadget | #PB_Window_SizeGadget
   ;
   UsePNGImageDecoder()  : UsePNGImageEncoder()
   UseJPEGImageDecoder() : UseJPEGImageEncoder()
   
   If OpenWindow(#WinMain,0,0,w,h,"Color Quantizing Image Editor"+"  |  "+#Version,flags) = 0
      ProcedureReturn #False
   EndIf
   
   SetWindowColor(#WinMain,$F6EAC3)
   CreateStatusBar(#status,WindowID(#WinMain))
   AddStatusBarField(#PB_Ignore) : AddStatusBarField(#PB_Ignore)
   ResizeWindow(#WinMain,#PB_Ignore,#PB_Ignore,w+130,h+StatusBarHeight(#status))
   WindowBounds(#WinMain,h,WindowHeight(#WinMain),#PB_Ignore,#PB_Ignore)
   
   CreatePopupMenu(#PopMenu)
      MenuItem(#MenuItem_Load,"load image")
      MenuItem(#MenuItem_Save,"save image")
      MenuItem(#MenuItem_Revert,"Revert")
      MenuItem(#MenuItem_New,"new image")
      MenuBar()
         MenuItem(#MenuItem_Set_Ref,"set work as reference")
      MenuBar()
         MenuItem(#MenuItem_Copy,"copy")
         MenuItem(#MenuItem_Paste,"paste")
         MenuItem(#MenuItem_Crop,"crop")
         MenuItem(#MenuItem_AddText,"add text")
      MenuBar()
         MenuItem(#MenuItem_Brightness,"Brightness")
         MenuItem(#MenuItem_Contrast,"Contrast")
      MenuBar()
         MenuItem(#MenuItem_Negative,"Negative")
         MenuItem(#MenuItem_Gray,"Gray Scale")
         MenuItem(#MenuItem_BW,"Black/White")
         MenuItem(#MenuItem_BW_Dither,"B/W dither")
      MenuBar()
         MenuItem(#MenuItem_Rot_R,"Rotate Right")
         MenuItem(#MenuItem_Rot_L,"Rotate Left")
         MenuItem(#MenuItem_Mirror,"Mirror")
         MenuItem(#MenuItem_Flip,"Flip")
   
   AddKeyboardShortcut(#WinMain, #PB_Shortcut_Return, #MenuItem_Enter)
   AddKeyboardShortcut(#WinMain, #PB_Shortcut_Left  , #MenuItem_LArrow)
   AddKeyboardShortcut(#WinMain, #PB_Shortcut_Right , #MenuItem_RArrow)
   AddKeyboardShortcut(#WinMain, #PB_Shortcut_Up    , #MenuItem_UArrow)
   AddKeyboardShortcut(#WinMain, #PB_Shortcut_Down  , #MenuItem_DArrow)
   
   CheckBoxGadget(#Chk_Hist,10,200,110,26,"Histogram")
   DisableGadget(#Chk_Hist,#True)
   
   CreateImage(#ImgArrow_L,38,23,32,$5F5F5F)
   StartDrawing(ImageOutput(#ImgArrow_L))
      LineXY(2,11,16,04,$FF00)
      LineXY(2,11,16,18,$FF00)
      LineXY(16,3,16,18,$FF00)
      FillArea(4,11,$FF00,$FF00)
      Box(17,8,18,6,$FF00)
   StopDrawing()
   
   CopyImage(#ImgArrow_L,#ImgArrow_R)
   MIRROR(#ImgArrow_R)
   
   ButtonImageGadget(#Btn_Undo,15,5,44,29,ImageID(#ImgArrow_L))
   GadgetToolTip(#Btn_Undo,"Undo")
   DisableGadget(#Btn_Undo,#True)
   
   
   ButtonImageGadget(#Btn_Redo,70,5,44,29,ImageID(#ImgArrow_R))
   GadgetToolTip(#Btn_Redo,"Redo")
   DisableGadget(#Btn_Redo,#True)
   
   FrameGadget(#PB_Any,5,230,115,50,"Unique colors")
      TextGadget(#cc,10,250,105,20,"")
   
   FrameGadget(#PB_Any,05,285,115,75,"View")
   OptionGadget(#Opt_ViewRef,10,305,100,20,"Reference")
   OptionGadget(#Opt_ViewWrk,10,330,100,20,"Work")
   SetGadgetState(#Opt_ViewRef,#True)
   
   FrameGadget(#PB_Any,5,40,115,155,"Quantize")
   ComboBoxGadget(#Combo_Qmode,10,60,105,30)
      AddGadgetItem(#Combo_Qmode,#Q_popular,        "Popularity")
         SetGadgetItemData(#Combo_Qmode,#Q_popular,32)
      AddGadgetItem(#Combo_Qmode,#Q_uniform,           "Uniform")
         SetGadgetItemData(#Combo_Qmode,#Q_uniform,3)
      AddGadgetItem(#Combo_Qmode,#Q_adaptive,         "Adaptive")
         SetGadgetItemData(#Combo_Qmode,#Q_adaptive,32)
      AddGadgetItem(#Combo_Qmode,#BMSK_Flat_8,          "flat_8")
      AddGadgetItem(#Combo_Qmode,#BMSK_flat_64,        "flat_64")
      AddGadgetItem(#Combo_Qmode,#BMSK_Flat_512,      "flat_512")
      AddGadgetItem(#Combo_Qmode,#BMSK_Flat_4096,    "flat_4096")
      AddGadgetItem(#Combo_Qmode,#BMSK_lemon_128,    "Lemon_128")
      AddGadgetItem(#Combo_Qmode,#BMSK_berry_128,    "Berry_128")
      AddGadgetItem(#Combo_Qmode,#BMSK_ocean_128,    "Ocean_128")
      AddGadgetItem(#Combo_Qmode,#Q_fixed_16,         "Fixed_16")
      AddGadgetItem(#Combo_Qmode,#Q_fixed_24,         "Fixed_24")
      AddGadgetItem(#Combo_Qmode,#Q_fixed_32,         "Fixed_32")
      AddGadgetItem(#Combo_Qmode,#Q_neuquqnt_256, "Neuquant_256")
   SetGadgetState(#Combo_Qmode,#Q_popular)
   
   CheckBoxGadget(#Chk_Dither,10,130,105,26,"Dither")
   ButtonGadget(#Btn_Apply,20,160,85,30,"Apply")
   
   SetGadgetColor(TextGadget(#PB_Any,70,95,35,25,"Limit"),#PB_Gadget_BackColor,$F6EAC3)
   SpinGadget(#Q_limit,10,100,55,25,2,256)
   SetGadgetState(#Q_limit,32)
   SetGadgetText(#Q_limit,"32")
   ScrollAreaGadget(#ScrollArea,125,5,w,h,w-4,h-14)
   CanvasGadget(#Canvas,0,0,w,h,#PB_Canvas_ClipMouse|#PB_Canvas_Keyboard)
   SetGadgetAttribute(#Canvas,#PB_Canvas_Cursor,#PB_Cursor_Cross)
   CloseGadgetList()
   
   BindGadgetEvent(#Canvas,@CANVAS_EVENT_HANDLER())
   
   CreateImage(#ImgMaster,w-4,h-14,32,$FFFFFF)
   StartDrawing(ImageOutput(#ImgMaster))
      DrawingMode(#PB_2DDrawing_AlphaBlend)
      For y = 0 To OutputHeight()-1
         For x = 0 To OutputWidth()-1
            r = (x ! y)
            g = r << 8
            b = g << 8 | $3F000000
            Plot(x, y, r | g | b)
         Next
      Next
      DrawText(10,10," RIGHT CLICK FOR MENU ")
   StopDrawing()
   
   GRAY(#ImgMaster)
   CopyImage(#ImgMaster,#ImgRef)
   CopyImage(#ImgMaster,#ImgWork)
   
   StartDrawing(CanvasOutput(#Canvas))
      DrawAlphaImage(ImageID(ImgActive),0,0)
   StopDrawing()
   
   ProcedureReturn #True
EndProcedure

Procedure EVENT_LOOP()
   Repeat
      Select WaitWindowEvent()
         Case #PB_Event_CloseWindow
            Select EventWindow()
               Case #WinMain : EndProgram = #True
               Case #WinHistogram : SetGadgetState(#Chk_Hist,#False) : DISPLAY_HISTOGRAM(-1)
               Case #WinContrast : CloseWindow(#WinContrast)
            EndSelect
         Case #PB_Event_SizeWindow : If EventWindow() = #WinMain : RESIZE() : EndIf
         Case #CanvasMouseDown     : SELECT_AREA(#Canvas)
         Case #CanvasRightClick    : DisplayPopupMenu(#PopMenu,WindowID(#WinMain))
         Case #CanvasCopy          : COPY()
         Case #CanvasPaste         : PASTE()
         Case #CanvasUndo          : UNDO()
         Case #CanvasRedo          : REDO()
         Case #PB_Event_MoveWindow : WINDOW_MOVED()
         Case #PB_Event_Menu
            Select EventMenu()
               Case #MenuItem_Revert     : REVERT()
               Case #MenuItem_Flip       : FLIP(ImgActive)        : UPDATE_CANVAS(ImgActive)
               Case #MenuItem_Mirror     : MIRROR(ImgActive)      : UPDATE_CANVAS(ImgActive)
               Case #MenuItem_Rot_L      : ROTATE_90(ImgActive,0) : UPDATE_CANVAS(ImgActive)
               Case #MenuItem_Rot_R      : ROTATE_90(ImgActive,1) : UPDATE_CANVAS(ImgActive)
               Case #MenuItem_Crop       : If CROP(ImgActive)     : UPDATE_CANVAS(ImgActive) : EndIf
               Case #MenuItem_BW         : BLACK_WHITE(ImgActive) : UPDATE_CANVAS(ImgActive)
               Case #MenuItem_BW_Dither  : BW_DITHER(ImgActive)   : UPDATE_CANVAS(ImgActive)
               Case #MenuItem_AddText    : ADD_TEXT()
               Case #MenuItem_Contrast   : ADJUST_CONTRAST()
               Case #MenuItem_Brightness : ADJUST_BRIGHTNESS()
               Case #MenuItem_Gray       : GRAY(ImgActive)        : UPDATE_CANVAS(ImgActive)
               Case #MenuItem_Negative   : NEGATIVE(ImgActive)    : UPDATE_CANVAS(ImgActive)
               Case #MenuItem_Set_Ref    : SET_AS_REFERENCE()
               Case #MenuItem_Copy       : COPY()
               Case #MenuItem_Paste      : PASTE()
               Case #MenuItem_Load       : LOAD_IMAGE()
               Case #MenuItem_Save       : SAVE_IMAGE()
               Case #MenuItem_New        : NEW_IMAGE()
               Case #MenuItem_Enter      : Q_LIMIT_ENTER_KEY()
            EndSelect
         Case #PB_Event_Gadget
            Select EventGadget()
               Case #Btn_Undo           : UNDO()
               Case #Btn_Redo           : REDO()
               Case #Chk_Hist           : CHOOSE_HISTOGRAM()
               Case #Opt_ViewRef        : TOG(#Opt_ViewRef)
               Case #Opt_ViewWrk        : TOG(#Opt_ViewWrk)
               Case #Btn_Apply          : EXECUTE_QUANTIZATION()
               Case #ChkRed To #ChkExEx : CHOOSE_HISTOGRAM()
               Case #Combo_Qmode        : CONFIG_Q_LIMIT()
               Case #Q_limit            : Q_LIMIT_ADJUST()
               Case #ImgGadPal          : EDIT_PALETTE_COLOR()
            EndSelect
      EndSelect
   Until EndProgram = #True
   
EndProcedure

Procedure UNDO()
   If (ImgActive = #ImgRef) And (UndoRedoStates & #UndoRef)
      If IsImage(#imgUndoRef)
         CopyImage(#ImgRef,#ImgTemp)
         CopyImage(#imgUndoRef,#ImgRef)
         CopyImage(#ImgTemp,#imgUndoRef)
         FreeImage(#ImgTemp)
         UndoRedoStates & %1100 | #RedoRef
         UPDATE_CANVAS(ImgActive)
         SET_UNDO_REDO_BUTTONS(ImgActive)
      EndIf
   ElseIf (ImgActive = #ImgWork) And (UndoRedoStates & #UndoWrk)
      If IsImage(#imgUndoWrk)
         CopyImage(#ImgWork,#ImgTemp)
         CopyImage(#imgUndoWrk,#ImgWork)
         CopyImage(#ImgTemp,#imgUndoWrk)
         FreeImage(#ImgTemp)
         UndoRedoStates & %0011 | #RedoWrk
         UPDATE_CANVAS(ImgActive)
         SET_UNDO_REDO_BUTTONS(ImgActive)
         SCAN_FOR_PALETTE()
      EndIf
   EndIf
EndProcedure

Procedure REDO()
   Protected iw = ImageWidth(ImgActive), ih = ImageHeight(ImgActive)
   
   If (ImgActive = #ImgRef) And (UndoRedoStates & #RedoRef)
      If IsImage(#imgUndoRef)
         CopyImage(#ImgRef,#ImgTemp)
         CopyImage(#imgUndoRef,#ImgRef)
         CopyImage(#ImgTemp,#imgUndoRef)
         FreeImage(#ImgTemp)
         UndoRedoStates & %1100 | #UndoRef
         UPDATE_CANVAS(ImgActive)
         SET_UNDO_REDO_BUTTONS(ImgActive)
      EndIf
   ElseIf (ImgActive = #ImgWork) And (UndoRedoStates & #RedoWrk)
      If IsImage(#imgUndoWrk)
         CopyImage(#ImgWork,#ImgTemp)
         CopyImage(#imgUndoWrk,#ImgWork)
         CopyImage(#ImgTemp,#imgUndoWrk)
         FreeImage(#ImgTemp)
         UndoRedoStates & %0011 | #UndoWrk
         UPDATE_CANVAS(ImgActive)
         SET_UNDO_REDO_BUTTONS(ImgActive)
         SCAN_FOR_PALETTE()
      EndIf
   EndIf
EndProcedure

Procedure CANVAS_EVENT_HANDLER()
   Select EventType()
      Case #PB_EventType_LeftButtonDown
         If GetGadgetAttribute(#Canvas,#PB_Canvas_Modifiers) = #PB_Canvas_Control
            ControlKey_Down = #True
         EndIf
         PostEvent(#CanvasMouseDown)
      Case #PB_EventType_LeftButtonUp   : PostEvent(#CanvasMouseUP)
      Case #PB_EventType_RightClick     : PostEvent(#CanvasRightClick)
      Case #PB_EventType_KeyUp
         If GetGadgetAttribute(#Canvas,#PB_Canvas_Modifiers) = #PB_Canvas_Control
            Select GetGadgetAttribute(#Canvas,#PB_Canvas_Key)
               Case #PB_Shortcut_C : PostEvent(#CanvasCopy)
               Case #PB_Shortcut_V : PostEvent(#CanvasPaste)
               Case #PB_Shortcut_Z : PostEvent(#CanvasUndo)
               Case #PB_Shortcut_Y : PostEvent(#CanvasRedo)
            EndSelect
         EndIf
   EndSelect
EndProcedure

Procedure COUNT_COLORS(image.i)
   ; returns the number of unique colors in an image (24 bit)
   ; stole this from wilbert
   Protected.i x, y, max_x, max_y, c, count, m
   Dim m.a($1FFFFF)
   StartDrawing(ImageOutput(image))
      max_x = ImageWidth(image) - 1
      max_y = ImageHeight(image) - 1
      For y = 0 To max_y
         For x = 0 To max_x
            c = Point(x, y) & $FFFFFF
            If m(c >> 3) & 1 << (c & 7) = 0
               m(c >> 3) | 1 << (c & 7)
               count + 1
            EndIf
         Next
      Next
   StopDrawing()
   ProcedureReturn count
EndProcedure

Procedure MN(n)
   ; returns magic number used for uniform color quantization
   ;n sets the number of quantization bins as #bins = n*n*n
   Protected div = Int(Round($FF/(n-0.5),#PB_Round_Nearest)) << 8
   Protected mlt = Int($FF/(n-1)), result
   result = div | mlt ;: Debug Hex(result)
   ProcedureReturn result
EndProcedure

Procedure SHOW_PALETTE(Array palette(1))
   ; draw the palette window
   
   Static flags = #PB_Window_SystemMenu | #PB_Window_Tool
   Protected c, i, x, y, columns, Xmax, Ymax, blockSize
   Protected Imax = ArraySize(palette())
   
   If IsWindow(#WinPalette) = 0
      x = WindowX(#WinMain,#PB_Window_InnerCoordinate) + 4
      y = WindowY(#WinMain,#PB_Window_InnerCoordinate) + GadgetY(#Opt_ViewWrk) + 40
      
      OpenWindow(#WinPalette,x,y,110,220,"",flags,WindowID(#WinMain))
   EndIf
   
   If Imax > 287 ; (12 * 24) - 1
      Imax = 287
      SetWindowTitle(#WinPalette,"Palette = " + "over!")
   Else
      SetWindowTitle(#WinPalette,"Palette = " + Str(Imax+1))
   EndIf
   
   columns = Round(Sqr((Imax+1)/2),#PB_Round_Up)
   blockSize = 112 / columns
   Xmax = columns * blockSize
   Ymax = Xmax * 2
   
   If IsImage(#imgPalette) = 0
      CreateImage(#imgPalette,Xmax,Ymax,24,0)
   Else
      ResizeImage(#imgPalette,Xmax,Ymax)
   EndIf
   
   StartDrawing(ImageOutput(#imgPalette))
      Xmax - 1 : Ymax - 1
      For y = 0 To Ymax
         For x = 0 To Xmax
            c = (x ! y)
            c = (c << 16) | (c << 8) | c
            Plot(x, y, c | $C0C0C0)
         Next
      Next
      X = 0 : Y = 0
      For I = 0 To Imax
         Box(X, Y, blockSize, blockSize, palette(I))
         X + blockSize
         If X > Xmax : X = 0 : Y + blockSize : EndIf
      Next
   StopDrawing()

   If IsGadget(#ImgGadPal) = 0
      ImageGadget(#ImgGadPal,0,0,ImageWidth(#imgPalette),ImageHeight(#imgPalette),ImageID(#imgPalette))
   Else
      SetGadgetState(#ImgGadPal,ImageID(#imgPalette))
   EndIf
   
EndProcedure

Procedure EDIT_PALETTE_COLOR()
   Static flag = #PB_MessageRequester_YesNo
   Protected c, i, x, y, MaxI, answer, newColor
   
   Macro ScanAndReplace(old, new)
      StartDrawing(ImageOutput(#ImgWork))
         y = 0 : While y < OutputHeight()
            x = 0 : While x < OutputWidth()
               If Point(x,y) = old : Plot(x,y,new) : EndIf
            x + 1 : Wend
         y + 1 : Wend
      StopDrawing()
      
      UPDATE_CANVAS(#ImgWork)
   EndMacro
   
   If EventType() = #PB_EventType_LeftClick
      x = WindowMouseX(#WinPalette) - GadgetX(#ImgGadPal)
      y = WindowMouseY(#WinPalette) - GadgetY(#ImgGadPal)
      StartDrawing(ImageOutput(#imgPalette))
         c = Point(x,y)
      StopDrawing()
      newColor = ColorRequester(c)
      If newColor <> -1
         CopyImage(#ImgWork,#ImgTemp)
         ScanAndReplace(c, newColor)
         
         answer = MessageRequester("","Do you wish to keep this color?",flag)
         
         If answer = #PB_MessageRequester_No
            ScanAndReplace(newColor, c)
         Else ; update the palette
            MaxI = ArraySize(WK_Pal())
            For i = 0 To MaxI
               If WK_Pal(i) = c : WK_Pal(i) = newColor : Break : EndIf
            Next
            SHOW_PALETTE(WK_Pal())
            CopyImage(#ImgTemp,#imgUndoWrk)
            UndoRedoStates & %0011 | #UndoWrk
            SET_UNDO_REDO_BUTTONS(#ImgWork)
         EndIf
         
         FreeImage(#ImgTemp)
      EndIf
   EndIf
   
EndProcedure

Procedure SCAN_FOR_PALETTE()
   ; obtain the palette of an image or selection
   Static NewMap Pmap.i(1024)
   Static Dim Palette(0)
   Protected c,i,x,y,Xmax,Ymax,image
   
   With selection
      If \sFlag
         GrabImage(ImgActive,#ImgSel,\left,\top,\right-\left,\bottom-\top)
         image = #ImgSel
      Else
         image = ImgActive
      EndIf
   EndWith
   
   Xmax = ImageWidth(image)-1
   Ymax = ImageHeight(image)-1
   
   StartDrawing(ImageOutput(image))
      For y = 0 To Ymax
         For x = 0 To Xmax
            c = Point(x,y)
               If MapSize(Pmap()) > 288
                  Break 2
               EndIf
            Pmap(Str(c)) = c
         Next
      Next
   StopDrawing()

   ReDim palette(MapSize(Pmap())-1)
   
   i = 0
   ForEach Pmap()
      palette(i) = Pmap()
      i + 1
   Next
   
   ClearMap(Pmap())
   
   If selection\sFlag = 0 : CopyArray(palette(),WK_Pal()) : EndIf
   
   SHOW_PALETTE(palette())
EndProcedure

Procedure TOG(option)
   ; switch the image displayed on the canvas
   Select option
      Case #Opt_ViewRef
         SetGadgetState(#Opt_ViewRef,#True)
         If selection\sFlag = #True And ImgActive = #ImgWork
            CANCEL_SELECTION() : SCAN_FOR_PALETTE()
         EndIf
         If IsWindow(#WinPalette) : HideWindow(#WinPalette,#True) : EndIf
         
         UPDATE_CANVAS(#ImgRef)
         SET_UNDO_REDO_BUTTONS(#ImgRef)
      Case #Opt_ViewWrk
         SetGadgetState(#Opt_ViewWrk,#True)
         If IsWindow(#WinPalette)
            HideWindow(#WinPalette,#False,#PB_Window_NoActivate)
         EndIf
         
         UPDATE_CANVAS(#ImgWork)
         SET_UNDO_REDO_BUTTONS(#ImgWork)
   EndSelect
EndProcedure

Procedure UPDATE_STATUS()
   ; update the selection information displayed in the status bar
   Protected text.s, lenX.i, lenY.i
   Shared selection.SelType
   
   With selection
      lenX = \right - \left + 1
      lenY = \bottom - \top + 1
      
      If \sFlag
         text = "Size = " + Str(lenX) + " x " + Str(lenY)
         text + " | x/y ratio = " + StrF(lenX / lenY,5)
         text + " | xLeft = "   + Str(\left)
         text + " | yTop = "    + Str(\top)
         text + " | xRight = "  + Str(\right)
         text + " | yBottom = " + Str(\bottom)
      EndIf
      
      StatusBarText(#status,1,text)
   EndWith
EndProcedure

Procedure COMPUTE_HISTOGRAM(image.i)
   Protected Xmax, Ymax, x, y, c, n, result
   
   If IsImage(image)
      With Hist
         For x = 0 To 255 : \R[x]=0 : \G[x]=0 : \B[x]=0 : Next x
         \maxR = 0 : \maxG = 0 : \maxB = 0
         
         StartDrawing(ImageOutput(image))
            Xmax = OutputWidth() - 1
            Ymax = OutputHeight() - 1
            
            For y = 0 To Ymax
               For x = 0 To Xmax
                  c = Point(x,y)
                  \R[c & $FF] + 1 : c >> 8
                  \G[c & $FF] + 1 : c >> 8
                  \B[c & $FF] + 1
               Next x
            Next y
         StopDrawing()
         
         Xmax = 255
         If IsGadget(#ChkExEx)
            If GetGadgetState(#ChkExEx) : n + 5 : Xmax - 5 : EndIf
         EndIf
         
         For x = n To Xmax
            If \R[x] > \maxR : \maxR = \R[x] : EndIf
            If \G[x] > \maxG : \maxG = \G[x] : EndIf
            If \B[x] > \maxB : \maxB = \B[x] : EndIf
         Next x
         
      EndWith
      
      result = #True
   EndIf
   
   ProcedureReturn result
EndProcedure

Procedure DISPLAY_HISTOGRAM(image.i)
   Static igad, flags = #PB_Window_WindowCentered | #PB_Window_SystemMenu
   Protected yScale.f, x, rc, gc, bc, h = 128 ;graph height
   
   If GetGadgetState(#Chk_Hist) = #True And image >= 0
      
      If IsWindow(#WinHistogram) = #False
         If OpenWindow(#WinHistogram,0,0,340,h+10,"Histogram",flags,WindowID(#WinMain)) = 0
            ProcedureReturn #False
         EndIf
         
         igad = ImageGadget(#PB_Any,5,5,256,h,0)
         CheckBoxGadget(#ChkRed, 270,010,50,25,"Red")
         CheckBoxGadget(#ChkGrn, 270,040,50,25,"Grn")
         CheckBoxGadget(#ChkBlu, 270,070,50,25,"Blu")
         CheckBoxGadget(#ChkExEx,270,100,50,25,"ExEx")
         GadgetToolTip(#ChkExEx,"Exclude Extremes")
         
         SetGadgetState(#ChkRed,#True)
         SetGadgetState(#ChkGrn,#True)
         SetGadgetState(#ChkBlu,#True)
         SetGadgetState(#ChkExEx,#True)
         
         If IsImage(#ImgHist) = #False
            CreateImage(#ImgHist,256,h)
         EndIf
      EndIf
      
      If COMPUTE_HISTOGRAM(image) = 0
         ProcedureReturn #False
      EndIf
      
      StartDrawing(ImageOutput(#ImgHist))
         With hist
            Box(0,0,256,h-1,0)
            DrawingMode(#PB_2DDrawing_XOr)
            
            If GetGadgetState(#ChkRed) : rc = $0000FF
               yScale = \maxR
            EndIf
            If GetGadgetState(#ChkGrn) : gc = $00FF00
               If \maxG > yScale : yScale = \maxG : EndIf
            EndIf
            If GetGadgetState(#ChkBlu) : bc = $FF0000
               If \maxB > yScale : yScale = \maxB : EndIf
            EndIf
            
            yScale = h / yScale : h - 1
            
            For x = 0 To 255
               LineXY(x, h, x, h - \R[x] * yScale, rc)
               LineXY(x, h, x, h - \G[x] * yScale, gc)
               LineXY(x, h, x, h - \B[x] * yScale, bc)
            Next x
         EndWith
      StopDrawing()
      
      SetGadgetState(igad,ImageID(#ImgHist))
   Else
      If IsWindow(#WinHistogram) <> #False
         FreeImage(#ImgHist)
         CloseWindow(#WinHistogram)
      EndIf
   EndIf
   
   ProcedureReturn #True
EndProcedure

Procedure CHOOSE_HISTOGRAM()
   ; determine source image for histogram
   With selection
      If \sFlag
         GrabImage(ImgActive,#ImgSel,\left,\top,\right-\left,\bottom-\top)
         DISPLAY_HISTOGRAM(#ImgSel)
      Else
         DISPLAY_HISTOGRAM(ImgActive)
      EndIf
   EndWith
EndProcedure

Procedure CANCEL_SELECTION()
   If selection\sFlag
      selection\active = #False
      selection\sFlag  = #False
      selection\top    = selection\bottom
      selection\left   = selection\right
      UPDATE_STATUS()
   EndIf
EndProcedure

Procedure SET_UNDO_REDO_BUTTONS(mode)
   If mode = #ImgRef
      DisableGadget(#Btn_Undo,Bool(Not(UndoRedoStates & #UndoRef)))
      DisableGadget(#Btn_Redo,Bool(Not(UndoRedoStates & #RedoRef)))
   ElseIf mode = #ImgWork
      DisableGadget(#Btn_Undo,Bool(Not(UndoRedoStates & #UndoWrk)))
      DisableGadget(#Btn_Redo,Bool(Not(UndoRedoStates & #RedoWrk)))
   EndIf
   SetActiveGadget(#Canvas)
EndProcedure

Procedure LOAD_IMAGE()
   Static m = #PB_Gadget_ContainerCoordinate, firstRun = #True
   Protected FileName$ = GetHomeDirectory() + "My Pictures\"
   Protected Pattern$ = "image (*.png, *.jpg, *.bmp)|*.png;*.jpg;*.bmp|image *.*|*.*"
   Protected text$
   
   File$ = OpenFileRequester("Select image to process", FileName$, Pattern$, 0)
   
   If File$
      SourceFileName$ = GetFilePart(File$,#PB_FileSystem_NoExtension)
      
      If LoadImage(#ImgTemp, File$)
         If firstRun = #True
            firstRun = #False
         Else
            PREPARE_UNDO(#ImgRef)
         EndIf
         
         CopyImage(#ImgTemp,#ImgRef)
         CopyImage(#ImgTemp,#ImgMaster)
         FreeImage(#ImgTemp)
         
         DisableGadget(#Chk_Hist,#False)
         
         TOG(#Opt_ViewRef)
      EndIf
   EndIf
EndProcedure

Procedure SAVE_IMAGE()
   Static YesNo = #PB_MessageRequester_YesNo, Yes = #PB_MessageRequester_Yes
   Protected Pattern$ = "image *.png|*.png|image *.jpg|*.jpg|image *.bmp|*.bmp"
   Protected text.s = SourceFileName$ + "_cq"+GetGadgetText(#CC)
   Protected p, img, cancel = #False
   If IsImage(#ImgMaster) = 0 : ProcedureReturn : EndIf
   Protected F$ = SaveFileRequester("Save image", text, Pattern$, 0)
   
   F$ = ReverseString(F$)
   p = FindString(F$,".")
   If p
      F$ = Right(F$,Len(F$)-p)
   EndIf
   F$ = ReverseString(F$)
   
   Select SelectedFilePattern()
      Case 0 : F$ + ".png"
      Case 1 : F$ + ".jpg"
      Case 2 : F$ + ".bmp"
   EndSelect
   
   If FileSize(F$) <> -1 ; file exists
      If MessageRequester("File Exists!", "Do you wish to overwrite?", YesNo) <> Yes
         cancel = #True
      EndIf
   EndIf
   
   If cancel = #False
      Select SelectedFilePattern()
         Case 0 : SaveImage(ImgActive, F$, #PB_ImagePlugin_PNG)
         Case 1 : SaveImage(ImgActive, F$, #PB_ImagePlugin_JPEG)
         Case 2 : SaveImage(ImgActive, F$, #PB_ImagePlugin_BMP)
      EndSelect
   EndIf
EndProcedure

Procedure REVERT()
   ; restore original image
   Protected text$
   
   If UndoRedoStates & %0011 = 0 : ProcedureReturn 0 : EndIf
   
   If IsImage(#ImgMaster)
      PREPARE_UNDO(#ImgRef)
      CopyImage(#ImgMaster,#ImgRef)
      
      TOG(#Opt_ViewRef)
   EndIf
EndProcedure

Procedure NEW_IMAGE()
   Static flags = #PB_Window_Tool|#PB_Window_WindowCentered|#PB_Window_SystemMenu
   Static iw = 800, ih = 600, color = $FFFFFF
   Protected c, imgGad, strGad_1, strGad_2, btnCancel, btnOK, imgbkgnd, ExitProcedure
   
   If OpenWindow(#WinNew,0,0,250,130,"New image",flags,WindowID(#WinMain))
      strGad_1 = StringGadget(#PB_Any,10,10,60,25,Str(iw),#PB_String_Numeric)
      strGad_2 = StringGadget(#PB_Any,10,40,60,25,Str(ih),#PB_String_Numeric)
      TextGadget(#PB_Any,75,15,50,25,"width")
      TextGadget(#PB_Any,75,40,50,25,"Height")
      TextGadget(#PB_Any,135,5,95,25,"Background color")
      btnOK = ButtonGadget(#PB_Any,190,100,50,25,"OK")
      btnCancel = ButtonGadget(#PB_Any,10,100,50,25,"Cancel")
      imgbkgnd = CreateImage(#PB_Any,90,45,24,color)
      
      If imgbkgnd
         imgGad = ImageGadget(#PB_Any,130,35,90,45,ImageID(imgbkgnd))
      EndIf
      
      Repeat
         Select WaitWindowEvent()
            Case #PB_Event_CloseWindow
               Select EventWindow()
                  Case #WinNew  : ExitProcedure = #True
                  Case #WinMain : ExitProcedure = #True : EndProgram = #True
               EndSelect
            Case #PB_Event_Gadget
               Select EventGadget()
                  Case imgGad
                     If EventType() = #PB_EventType_LeftClick
                        c = ColorRequester(color)
                        If c >= 0 : color = c
                           StartDrawing(ImageOutput(imgbkgnd))
                              FillArea(0,0,c!1,c)
                           StopDrawing()
                           SetGadgetState(imgGad,ImageID(imgbkgnd))
                        EndIf
                     EndIf
                  Case strGad_1
                     If EventType() = #PB_EventType_Change
                        iw = Val(GetGadgetText(strGad_1))
                     EndIf
                  Case strGad_2
                     If EventType() = #PB_EventType_Change
                        ih = Val(GetGadgetText(strGad_2))
                     EndIf
                  Case btnOK
                     If iw < 16 : iw = 16 : EndIf
                     If ih < 16 : ih = 16 : EndIf
                     If iw > 4096 : iw = 4096 : EndIf
                     If ih > 4096 : ih = 4096 : EndIf
                     PREPARE_UNDO(ImgActive)
                     CreateImage(ImgActive,iw,ih,24,color)
                     UPDATE_CANVAS(ImgActive)
                     ExitProcedure = #True
                  Case btnCancel
                     ExitProcedure = #True
               EndSelect
         EndSelect
      Until ExitProcedure = #True
      
      CloseWindow(#WinNew)
   EndIf
EndProcedure

Procedure COPY()
   ; copy active image or selection to clipboard
   Protected img
   
   With selection
      If \sFlag
         img = GrabImage(ImgActive,#PB_Any,\left,\top,\right-\left,\bottom-\top)
         If img
            SetClipboardImage(img)
            UPDATE_CANVAS(ImgActive)
         EndIf
      Else
         SetClipboardImage(ImgActive)
      EndIf
   EndWith

EndProcedure

Procedure PASTE()
   ; display the clipboard contents as a floating section that
   ; follows the mouse cursor.  Paste when the left mouse button
   ; is clicked.  Press escape key to abort paste. Use the arrow
   ; keys to rotate and flip the floating section.
   
   Protected x,y,cw,ch,iw,ih,xx,yy,ExitProcedure
   
   Macro DrawFloatingSection(rotated)
      If rotated = #True
         iw = ImageWidth(#ImgTemp)
         ih = ImageHeight(#ImgTemp)
         x + xx    : y + yy
         xx = iw/2 : yy = ih/2
         x - xx    : y - yy
      EndIf
      StartDrawing(CanvasOutput(#Canvas))
         DrawImage(ImageID(ImgActive),0,0)
         DrawImage(ImageID(#ImgTemp),x,y)
      StopDrawing()
   EndMacro
   
   If GetClipboardImage(#ImgTemp)
      cw = ImageWidth(ImgActive)
      ch = ImageHeight(ImgActive)
      iw = ImageWidth(#ImgTemp)
      ih = ImageHeight(#ImgTemp)
      xx = iw/2
      yy = ih/2
      
      If iw > cw Or ih > ch
         If iw > cw : iw = cw : EndIf
         If ih > ch : ih = ch : EndIf
         GrabImage(#ImgTemp,#ImgTemp,0,0,iw,ih)
      EndIf
      
      x = cw/2 - xx
      y = ch/2 - yy
      DrawFloatingSection(#False)
      
      Repeat
         Select WaitWindowEvent()
            Case #PB_Event_CloseWindow
               If EventWindow() = #WinMain
                  ExitProcedure = #True
                  EndProgram = #True
               EndIf
            Case #CanvasMouseDown
               PREPARE_UNDO(ImgActive)
               StartDrawing(ImageOutput(ImgActive))
                  DrawImage(ImageID(#ImgTemp),x,y)
               StopDrawing()
               If ImgActive = #ImgWork : SCAN_FOR_PALETTE() : EndIf
               ExitProcedure = #True
            Case #PB_Event_Menu
               Select EventMenu()
                  Case #MenuItem_LArrow : ROTATE_90(#ImgTemp,0) : DrawFloatingSection(#True)
                  Case #MenuItem_RArrow : ROTATE_90(#ImgTemp,1) : DrawFloatingSection(#True)
                  Case #MenuItem_UArrow : MIRROR(#ImgTemp)      : DrawFloatingSection(#False)
                  Case #MenuItem_DArrow : FLIP(#ImgTemp)        : DrawFloatingSection(#False)
               EndSelect
            Case #PB_Event_Gadget
               If EventGadget() = #Canvas
                  Select EventType()
                     Case #PB_EventType_MouseMove
                        x = GetGadgetAttribute(#Canvas,#PB_Canvas_MouseX) - xx
                        y = GetGadgetAttribute(#Canvas,#PB_Canvas_MouseY) - yy
                        If x + iw > cw : x = cw - iw : EndIf
                        If y + ih > ch : y = ch - ih : EndIf
                        If x < 0 : x = 0 : EndIf
                        If y < 0 : y = 0 : EndIf
                        DrawFloatingSection(#False)
                     Default
                        If #PB_Shortcut_Escape = GetGadgetAttribute(#Canvas,#PB_Canvas_Key)
                           ExitProcedure = #True
                        EndIf
                  EndSelect
               EndIf
               
         EndSelect
      Until ExitProcedure = #True
      
      FreeImage(#ImgTemp)
      UPDATE_CANVAS(ImgActive)
   EndIf
EndProcedure

Procedure RESIZE()
   ; main window has been resized
   Protected w,h,x,y
   
   h = WindowHeight(#WinMain) - 10 - StatusBarHeight(#status)
   w = WindowWidth(#WinMain) - 130
   ResizeGadget(#ScrollArea,#PB_Ignore,#PB_Ignore,w,h)
   
   w = ImageWidth(ImgActive)
   h = ImageHeight(ImgActive)
   ResizeGadget(#Canvas,#PB_Ignore,#PB_Ignore,w,h)
   SET_SELECTION_BOUNDARIES(#Canvas)
   
   If IsWindow(#WinPalette)
      x = WindowX(#WinMain,#PB_Window_InnerCoordinate) + 4
      y = WindowY(#WinMain,#PB_Window_InnerCoordinate) + GadgetY(#Opt_ViewWrk) + 40
      ResizeWindow(#WinPalette,x,y,#PB_Ignore,#PB_Ignore)
   EndIf
   
EndProcedure

Procedure ADD_TEXT()
   Static flags = #PB_Window_Tool|#PB_Window_WindowCentered|#PB_Window_SystemMenu
   Static text$ = "Enter text here."
   Static fontName.s = "Arial"
   Static fontSize = 10
   Static fontColor = 0
   Static bakColor = $FFFFFF
   Static fontStyle = 0
   Static style$ = "Normal"
   Static NewMap style()
   Static firstRun = #True
   Static fontNum, img_Fcolor, img_Bcolor
   Protected btnCancel, btnOK, str_name, str_size, str_style, edt_text
   Protected imgGad_Fcolor, imgGad_Bcolor, ExitProcedure
   Protected c, w, h, L, m, n, x, y, t$
   
   Macro SelectFont()
      If FontRequester(fontName,fontSize,#PB_FontRequester_Effects,fontColor,fontStyle)
         fontName  = SelectedFontName()
         SetGadgetText(str_name, fontName)
         fontColor = SelectedFontColor()
            StartDrawing(ImageOutput(img_Fcolor))
               FillArea(0, 0, fontColor!1, fontColor)
            StopDrawing()
            SetGadgetState(imgGad_Fcolor,ImageID(img_Fcolor))
            SetGadgetColor(edt_text,#PB_Gadget_FrontColor,fontColor)
            fontSize  = SelectedFontSize()
            SetGadgetText(str_size,Str(fontSize))
         fontStyle = SelectedFontStyle()
         If fontStyle = 0 : style$ = "Normal"
         Else
            style$ = ""
            ForEach style()
               If style() & fontStyle
                  style$ + MapKey(style())
               EndIf
            Next
         EndIf
         SetGadgetText(str_style, style$)
         FreeFont(fontNum)
         fontNum = LoadFont(#PB_Any,fontName,fontSize,fontStyle)
      EndIf
   EndMacro
   
   If firstRun
      firstRun = #False
      style("Bold ")         = #PB_Font_Bold
      style("High Quality ") = #PB_Font_HighQuality
      style("Italic ")       = #PB_Font_Italic
      style("Strike Out ")   = #PB_Font_StrikeOut
      style("Underline ")    = #PB_Font_Underline
      fontNum = LoadFont(#PB_Any,fontName,fontSize,fontStyle)
      img_Fcolor = CreateImage(#PB_Any,25,25,24,fontColor)
      img_Bcolor = CreateImage(#PB_Any,25,25,24,bakColor)
   EndIf
   
   If OpenWindow(#WinText,0,0,280,270,"Add text",flags,WindowID(#WinMain))
      btnCancel = ButtonGadget(#PB_Any,10,240,60,30,"Cancel")
      btnOK     = ButtonGadget(#PB_Any,210,240,60,30,"OK")
      
      edt_text  = EditorGadget(#PB_Any,10,130,260,100)
      SetGadgetFont(edt_text,FontID(fontNum))
      SetGadgetColor(edt_text,#PB_Gadget_FrontColor,fontColor)
      SetGadgetColor(edt_text,#PB_Gadget_BackColor,bakColor)
      SetGadgetText(edt_text,text$)
      
      TextGadget(#PB_Any,10,5,120,25,"Font name")
      str_name = StringGadget(#PB_Any,10,30,170,25,fontName,#PB_String_ReadOnly)
      
      TextGadget(#PB_Any,190,5,60,25,"Size")
      str_size = StringGadget(#PB_Any,190,30,40,25,Str(fontSize),#PB_String_ReadOnly)
      
      TextGadget(#PB_Any,10,70,60,25,"Style")
      str_style = StringGadget(#PB_Any,10,95,170,25,style$,#PB_String_ReadOnly)
      
      TextGadget(#PB_Any,190,70,50,25,"color")
      imgGad_Fcolor = ImageGadget(#PB_Any,190,95,25,25,ImageID(img_Fcolor))
      imgGad_Bcolor = ImageGadget(#PB_Any,220,95,25,25,ImageID(img_Bcolor))
      
      Repeat
         Select WaitWindowEvent()
            Case #PB_Event_CloseWindow
               Select EventWindow()
                  Case #WinText : ExitProcedure = 1
                  Case #WinMain : ExitProcedure = 1 : EndProgram = #True
               EndSelect
            Case #PB_Event_Gadget
               Select EventGadget()
                  Case btnCancel : ExitProcedure = 1
                     
                  Case btnOK
                     ; determine the required image size to contain text
                     text$ = GetGadgetText(edt_text)
                     CreateImage(#ImgTemp, 5, 5, 24, bakColor) ; will resize later
                     StartDrawing(ImageOutput(#ImgTemp))
                        DrawingFont(FontID(fontNum))
                        n = CountGadgetItems(edt_text)
                        h = TextHeight("A1g") * n ; actual text here not important
                        w = TextWidth(" ") ; use this value if text box is empty
                        For m = 0 To n - 1 ; examine each line of text
                           L = TextWidth(GetGadgetItemText(edt_text,m))
                           If L > w : w = L : EndIf ; find length of longest line
                        Next m
                     StopDrawing()
                     
                     ; now we know the proper size
                     ResizeImage(#ImgTemp, w + 10, h + 10) ; 5 pixel border
                     
                     ; draw all of the text on the temporary image
                     StartDrawing(ImageOutput(#ImgTemp))
                        DrawingFont(FontID(fontNum))
                        x = 5 : y = 5
                        For m = 0 To n - 1
                           t$ = GetGadgetItemText(edt_text, m)
                           If t$ = "" : t$ = " " : EndIf
                           DrawText(x, y, t$, fontColor, bakColor)
                           y + TextHeight(t$)
                        Next m
                     StopDrawing()
                     
                     SetClipboardImage(#ImgTemp) ; <-- we are finished
                     FreeImage(#ImgTemp)
                     ExitProcedure = 2 ; this value will activate paste on exit
                     
                  Case str_name, str_size, str_style
                     ; the active font will be updated
                     If EventType() = #PB_EventType_Focus
                        SelectFont() ; macro
                        While WindowEvent() : Wend
                        SetActiveGadget(edt_text)
                        SetGadgetFont(edt_text,FontID(fontNum))
                     EndIf
                     
                  Case imgGad_Fcolor
                     ; the text front color will be chosen
                     If EventType() = #PB_EventType_LeftClick
                        c = ColorRequester(fontColor)
                        If c >= 0 : fontColor = c
                           StartDrawing(ImageOutput(img_Fcolor))
                              FillArea(0,0,c!1,c)
                           StopDrawing()
                           SetGadgetState(imgGad_Fcolor,ImageID(img_Fcolor))
                           SetGadgetColor(edt_text,#PB_Gadget_FrontColor,fontColor)
                        EndIf
                     EndIf
                     
                  Case imgGad_Bcolor
                     ; the text back color will be chosen
                     If EventType() = #PB_EventType_LeftClick
                        c = ColorRequester(bakColor)
                        If c >= 0 : bakColor = c
                           StartDrawing(ImageOutput(img_Bcolor))
                              FillArea(0,0,c!1,c)
                           StopDrawing()
                           SetGadgetState(imgGad_Bcolor,ImageID(img_Bcolor))
                           SetGadgetColor(edt_text,#PB_Gadget_BackColor,bakColor)
                        EndIf
                     EndIf
               EndSelect
         EndSelect
      Until ExitProcedure <> 0
      
      CloseWindow(#WinText)
      
      If ExitProcedure = 2 : PASTE() : EndIf
   EndIf
EndProcedure

Procedure UPDATE_CANVAS(image)
   Protected iw, ih, cw, ch, text$
   
   If IsImage(image)
      cw = GadgetWidth(#Canvas, #PB_Gadget_ActualSize)
      ch = GadgetHeight(#Canvas,#PB_Gadget_ActualSize)
      iw = ImageWidth(image)
      ih = ImageHeight(image)
      
      If cw <> iw Or ch <> ih
         ResizeGadget(#Canvas,0,0,ImageWidth(image),ImageHeight(image))
         SetGadgetAttribute(#ScrollArea,#PB_ScrollArea_InnerWidth,ImageWidth(image))
         SetGadgetAttribute(#ScrollArea,#PB_ScrollArea_InnerHeight,ImageHeight(image))
         SET_SELECTION_BOUNDARIES(#Canvas)
         
         text$ = File$ + "  |  (" + iw + " x " + ih + ")"
         StatusBarText(#status, 0, text$)
      EndIf
      
      ImgActive = image
      StartDrawing(CanvasOutput(#Canvas))
         DrawAlphaImage(ImageID(image),0,0)
      StopDrawing()
      
      If selection\sFlag = #True
         CANCEL_SELECTION()
      EndIf
      
      SetGadgetText(#cc,Str(COUNT_COLORS(image)))
      
      If GetGadgetState(#Chk_Hist) : DISPLAY_HISTOGRAM(image) : EndIf
   EndIf
EndProcedure

Procedure PREPARE_UNDO(image)
   If image = #ImgRef
      CopyImage(image,#imgUndoRef)
      UndoRedoStates & %1100 | #UndoRef
      SET_UNDO_REDO_BUTTONS(#ImgRef)
   ElseIf image = #ImgWork
      CopyImage(image,#imgUndoWrk)
      UndoRedoStates & %0011 | #UndoWrk
      SET_UNDO_REDO_BUTTONS(#ImgWork)
   EndIf
EndProcedure

Procedure FLIP(image)
   ; top becomes bottom, bottom becomes top
   Protected d,x,y,yp,Xmax,Ymax,Ylim
   
   If IsImage(image)
      PREPARE_UNDO(image)
      
      StartDrawing(ImageOutput(image))
         Xmax = OutputWidth() - 1
         Ymax = OutputHeight() - 1
         Ylim = OutputHeight() / 2 - 1
         
         For x = 0 To Xmax
            For y = 0 To Ylim
               yp = Ymax - y
               d = Point(x,y)
               Plot(x,y,Point(x,yp))
               Plot(x,yp,d)
            Next y
         Next x
         
      StopDrawing()
   EndIf
EndProcedure

Procedure MIRROR(image)
   ; left becomes right, right becomes left
   Protected d,x,y,xp,Xmax,Ymax,Xlim
   
   If IsImage(image)
      PREPARE_UNDO(image)
      
      StartDrawing(ImageOutput(image))
         Xmax = OutputWidth() - 1
         Xlim = OutputWidth() / 2 - 1
         Ymax = OutputHeight() - 1
         
         For y = 0 To Ymax
            For x = 0 To Xlim
               xp = Xmax - x
               d = Point(x,y)
               Plot(x,y,Point(xp,y))
               Plot(xp,y,d)
            Next x
         Next y
         
      StopDrawing()
   EndIf
EndProcedure

Procedure ROTATE_90(image, dir)
   ; rotate image +/- 90 deg.
   ; 'image' is the number of the image to rotate
   ; if 'dir' = 0 then rotate left else rotate right
   
   Protected a,b,c,e,f,h,s,w,x,y,ym,xm,tempImg,depth
   
   If IsImage(image) = 0 : ProcedureReturn 0 : EndIf
   
   StartDrawing(ImageOutput(image))
      w = OutputWidth()
      h = OutputHeight()
      f = DrawingBufferPixelFormat() & $7F
   StopDrawing()
      
      If f = #PB_PixelFormat_32Bits_RGB Or f = #PB_PixelFormat_32Bits_BGR
         depth = 32
      ElseIf f = #PB_PixelFormat_24Bits_RGB Or f = #PB_PixelFormat_24Bits_BGR
         depth = 24
      Else
         ProcedureReturn 0
      EndIf
   
   If w > h : s = w : Else : s = h : EndIf ; find the largest dimension
   
   tempImg = CreateImage(#PB_Any,s,s,depth) ; make a square working area
   
   StartDrawing(ImageOutput(tempImg))
      If depth = 32 : DrawingMode(#PB_2DDrawing_AllChannels) : EndIf
      
      DrawImage(ImageID(image),0,0)
      
      ym = s/2-1 ; max y loop value
      xm = s/2-(s!1&1) ; max x value, subtract 1 if 's' is even
      s-1
      
      If dir <> 0 ; rotate right
         For y = 0 To ym
            For x = 0 To xm
               e = Point(x,y)
               a = s-x : Plot(x,y,Point(y,a))
               b = s-y : Plot(y,a,Point(a,b))
               c = s-a : Plot(a,b,Point(b,c))
               Plot(b,c,e)
            Next x
         Next y
      Else ; rotate left
         For y = 0 To ym
            For x = 0 To xm
               e = Point(x,y)
               a = s-y : Plot(x,y,Point(a,x))
               b = s-x : Plot(a,x,Point(b,a))
               c = s-a : Plot(b,a,Point(c,b))
               Plot(c,b,e)
            Next x
         Next y
      EndIf
      
   StopDrawing()
   
   PREPARE_UNDO(image)
   
   If dir <> 0
      GrabImage(tempImg,image,s-h+1,0,h,w) ; right
   Else
      GrabImage(tempImg,image,0,s-w+1,h,w) ; left
   EndIf
   
   FreeImage(tempImg)
   
   ProcedureReturn 1
EndProcedure

Procedure NEGATIVE(image)
   ; produces a negative color image
   ; 0 --> 255, 1 --> 254, 2 --> 253 ...
   Protected x,y,Xmax,Ymax
   
   If IsImage(image)
      PREPARE_UNDO(image)
      
      StartDrawing(ImageOutput(image))
         Xmax = OutputWidth() - 1
         Ymax = OutputHeight() - 1
         
         For y = 0 To Ymax
            For x = 0 To Xmax
               Plot(x,y, Point(x,y) ! $FFFFFF)
            Next x
         Next y
         
      StopDrawing()
      
      If ImgActive = #ImgWork : SCAN_FOR_PALETTE() : EndIf
   EndIf
EndProcedure

Procedure BLACK_WHITE(image)
   ; convert image to 2 colors only
   ; the colors are black and white by default
   ; colors are weighted as follows for luminence conversion:
   ; R=0.286, G=0.571, B=0.143
   
   Static kb=$FF0000, kg=$00FF00, kr=$0000FF
   Static c1 = 0, c2 = $FFFFFF ; default colors
   Protected x,y,d,Ymax,Xmax,lum
   
   If IsImage(image)
      PREPARE_UNDO(image)
      
      StartDrawing(ImageOutput(image))
         Xmax = OutputWidth() - 1
         Ymax = OutputHeight() - 1
         
         For y = 0 To Ymax
            For x = 0 To Xmax
               d = Point(x,y)
               lum = (d & kr)<<1 + (d & kg) >> 6 + (d & kb) >> 16
               lum / 7
               If lum  < $80 ; threshold
                  Plot(x,y,c1)
               Else
                  Plot(x,y,c2)
               EndIf
            Next x
         Next y
      StopDrawing()
      
      If ImgActive = #ImgWork : SCAN_FOR_PALETTE() : EndIf
   EndIf
EndProcedure

Procedure BW_DITHER(image)
   ; Sierra Lite dither algorithm
   ; coded in PureBasic by Wilbert
   ; modified to remove threshold control by BasicallyPure
   ; image is converted to two colors (black & white by default)
   
   ;Protected errorMul = GetGadgetState(#TrackBar_Threshold) ; option
   Static c1 = 0, c2 = $FFFFFF ; default colors
   Protected x, y, w, h, v, error
   
   If IsImage(image)
      PREPARE_UNDO(image)
      
      StartDrawing(ImageOutput(image))
         w = OutputWidth() - 1
         h = OutputHeight() - 1
         Dim errorBuffer(w+1)
         
         For y = 0 To h
            error = 0
            For x = 0 To w
               error + errorBuffer(x + 1)
               v = Point(x, y)
               v = (Red(v) * 77 + Green(v) * 150 + Blue(v) * 29) >> 8 + error
               If v >= 128
                  Plot(x, y, c2)
                  If v > 255
                     error = 0
                  Else
                     error = v - 255
                  EndIf
               Else
                  Plot(x, y, c1)
                  If v < 0
                     error = 0
                  Else
                     error = v
                  EndIf
               EndIf
               error >> 2 ;error = (error * errorMul) >> 9 ;option
               errorBuffer(x + 1) = error >> 1 
               errorBuffer(x) + errorBuffer(x + 1)
            Next x
         Next y
      StopDrawing()
      
      If ImgActive = #ImgWork : SCAN_FOR_PALETTE() : EndIf
   EndIf
EndProcedure

Procedure ADJUST_CONTRAST()
   Static flags = #PB_Window_Tool|#PB_Window_WindowCentered|#PB_Window_SystemMenu
   Protected.i Trkbar, TxtGad, BtnReset, BtnApply, BtnCancel, BtnOK
   Protected.i r, g, b, x, y, xm, ym, ExitProcedure
   Protected.f contrast
   
   If IsWindow(#WinContrast) : ProcedureReturn 0 : EndIf
   
   If OpenWindow(#WinContrast,0,0,250,90,"Contrast Adjust",flags,WindowID(#WinMain))
      TrkBar = TrackBarGadget(#PB_Any,4,10,200,25,0,200)
      SetGadgetState(TrkBar,100)
      TxtGad    = TextGadget(#PB_Any,205,15,40,25,"0")
      BtnReset  = ButtonGadget(#PB_Any ,010,50,50,25,"Reset")
      BtnApply  = ButtonGadget(#PB_Any ,070,50,50,25,"Apply")
      BtnCancel = ButtonGadget(#PB_Any ,130,50,50,25,"Cancel")
      BtnOK     = ButtonGadget(#PB_Any ,190,50,50,25,"OK")
      
      Macro ConCalc(n) ; contrast calculation
         If n < 128
            n = n - contrast*(127 - n)
            If n < 0 : n = 0 : EndIf
         Else
            n = n + contrast*(n - 128)
            If n > 255 : n = 255 : EndIf
         EndIf
      EndMacro
      
      Macro ApplyContrast()
         CopyImage(ImgActive,#ImgTemp)
         StartDrawing(ImageOutput(#ImgTemp))
            contrast = ((GetGadgetState(TrkBar)-100)/100)
            ym = OutputHeight() - 1
            xm = OutputWidth() - 1
            For y = 0 To ym
               For x = 0 To xm
                  b = Point(x,y)
                  r = b & $FF : b >> 8
                  g = b & $FF : b >> 8
                  ConCalc(b) : ConCalc(g) : ConCalc(r)
                  Plot(x, y, b<<16 | g<<8 | r)
               Next x
            Next y
         StopDrawing()
         
         StartDrawing(CanvasOutput(#Canvas))
            DrawAlphaImage(ImageID(#ImgTemp),0,0)
         StopDrawing()
         
         If GetGadgetState(#Chk_Hist) : DISPLAY_HISTOGRAM(#ImgTemp) : EndIf
      EndMacro
      
      Macro CancelContrast()
         If IsImage(#ImgTemp) : FreeImage(#ImgTemp) : EndIf
         UPDATE_CANVAS(ImgActive)
         CloseWindow(#WinContrast)
         ExitProcedure = #True
      EndMacro
      
      Repeat
         Select WaitWindowEvent()
            Case #PB_Event_CloseWindow
               Select EventWindow()
                  Case #WinContrast : CancelContrast()
                  Case #WinMain    : ExitProcedure = #True : EndProgram = #True
               EndSelect
            Case #PB_Event_Gadget
               Select EventGadget()
                  Case Trkbar    : SetGadgetText(TxtGad,Str(GetGadgetState(Trkbar)-100))
                  Case BtnReset  : SetGadgetState(Trkbar,100) : SetGadgetText(TxtGad,"0")
                     ApplyContrast()
                  Case BtnCancel : CancelContrast()
                  Case BtnApply  : ApplyContrast()
                  Case #ChkRed To #ChkExEx : DISPLAY_HISTOGRAM(#ImgTemp)
                  Case BtnOK
                     If IsImage(#ImgTemp)
                        PREPARE_UNDO(ImgActive)
                        CopyImage(#ImgTemp,ImgActive)
                        FreeImage(#ImgTemp)
                        UPDATE_CANVAS(ImgActive)
                        If ImgActive = #ImgWork : SCAN_FOR_PALETTE() : EndIf
                     EndIf
                     CloseWindow(#WinContrast) : ExitProcedure = #True
               EndSelect
         EndSelect
      Until ExitProcedure = #True
   EndIf
EndProcedure

Procedure ADJUST_BRIGHTNESS()
   Static flags = #PB_Window_Tool|#PB_Window_WindowCentered|#PB_Window_SystemMenu
   Protected TrkBar, TxtGad, BtnReset, BtnApply, BtnCancel, BtnOK
   Protected Chk_R, Chk_G, Chk_B, Chk_OvrDrv, overdrive
   Protected.i r, g, b, x, y, xm, ym, rb, gb, bb, ExitProcedure
   Protected.f brightness
   
   If IsWindow(#WinBright) : ProcedureReturn 0 : EndIf
   
   If OpenWindow(#WinBright,0,0,300,120,"Brightness Adjust",flags,WindowID(#WinMain))
      TrkBar     = TrackBarGadget(#PB_Any,4,10,257,25,0,256)
         SetGadgetState(TrkBar,128)
      TxtGad     = TextGadget(#PB_Any,265,015,40,25,"0")
      Chk_R      = CheckBoxGadget(#PB_Any,010,45,40,25,"RED")
         SetGadgetState(Chk_R,#True)
      Chk_G      = CheckBoxGadget(#PB_Any,065,45,40,25,"GRN")
         SetGadgetState(Chk_G,#True)
      Chk_B      = CheckBoxGadget(#PB_Any,120,45,40,25,"BLU")
         SetGadgetState(Chk_B,#True)
      Chk_OvrDrv = CheckBoxGadget(#PB_Any,180,45,80,25,"OVERDRIVE")
         SetGadgetState(Chk_OvrDrv,#False)
      BtnReset   = ButtonGadget(#PB_Any ,015,85,60,25,"Reset")
      BtnApply   = ButtonGadget(#PB_Any ,085,85,60,25,"Apply")
      BtnCancel  = ButtonGadget(#PB_Any ,155,85,60,25,"Cancel")
      BtnOK      = ButtonGadget(#PB_Any ,225,85,60,25,"OK")
      
      Macro CancelBrightness()
         If IsImage(#ImgTemp) : FreeImage(#ImgTemp) : EndIf
         UPDATE_CANVAS(ImgActive)
         CloseWindow(#WinBright)
         ExitProcedure = #True
      EndMacro
      
      Macro Clip(value)
         If value > 255 : value = 255 : EndIf
         If value < 000 : value = 000 : EndIf
      EndMacro
      
      Macro ApplyBrightness()
         CopyImage(ImgActive,#ImgTemp)
         StartDrawing(ImageOutput(#ImgTemp))
            overdrive = GetGadgetState(Chk_OvrDrv)
            brightness = ((GetGadgetState(TrkBar)-128)/128.0)
            rb = brightness * 128 * GetGadgetState(Chk_R)
            gb = brightness * 128 * GetGadgetState(Chk_G)
            bb = brightness * 128 * GetGadgetState(Chk_B)
            ym = OutputHeight() - 1
            xm = OutputWidth() - 1
            
            For y = 0 To ym
               For x = 0 To xm
                  b = Point(x,y)
                  r = b & $FF : b >> 8
                  g = b & $FF : b >> 8
                  
                  b + bb : g + gb : r + rb
                  
                  If Not overdrive
                     Clip(b) : Clip(g) : clip(r)
                  EndIf
                  
                  Plot(x, y, b<<16 | g<<8 | r)
               Next x
            Next y
            
         StopDrawing()
         
         StartDrawing(CanvasOutput(#Canvas))
            DrawAlphaImage(ImageID(#ImgTemp),0,0)
         StopDrawing()
         
         If GetGadgetState(#Chk_Hist) : DISPLAY_HISTOGRAM(#ImgTemp) : EndIf
      EndMacro
      
      Repeat
         Select WaitWindowEvent()
            Case #PB_Event_CloseWindow
               Select EventWindow()
                  Case #WinBright : CancelBrightness()
                  Case #WinMain   : ExitProcedure = #True : EndProgram = #True
               EndSelect
            Case #PB_Event_Gadget
               Select EventGadget()
                  Case Trkbar    : SetGadgetText(TxtGad,StrF(GetGadgetState(Trkbar)-128))
                  Case BtnReset  : SetGadgetState(Trkbar,128) : SetGadgetText(TxtGad,"0")
                     ApplyBrightness()
                  Case BtnCancel : CancelBrightness()
                  Case BtnApply  : ApplyBrightness()
                  Case #ChkRed To #ChkExEx : DISPLAY_HISTOGRAM(#ImgTemp)
                  Case BtnOK
                     If IsImage(#ImgTemp)
                        PREPARE_UNDO(ImgActive)
                        CopyImage(#ImgTemp,ImgActive)
                        FreeImage(#ImgTemp)
                        UPDATE_CANVAS(ImgActive)
                        If ImgActive = #ImgWork : SCAN_FOR_PALETTE() : EndIf
                     EndIf
                     CloseWindow(#WinBright) : ExitProcedure = #True
               EndSelect
         EndSelect
      Until ExitProcedure = #True
   EndIf
   
EndProcedure

Procedure GRAY(image)
   ; reduce colors to 256 grayscale
   ; color weights are R=0.286, G=0.571, B=0.143
   
   Static kb=$FF0000, kg=$00FF00, kr=$0000FF
   Protected x,y,d,Ymax,Xmax,lum
   
   If IsImage(image)
      PREPARE_UNDO(image)
      
      StartDrawing(ImageOutput(image))
         Xmax = OutputWidth() - 1
         Ymax = OutputHeight() - 1
         
         For y = 0 To Ymax
            For x = 0 To Xmax
               d = Point(x,y)
               lum = (d & kr)<<1 + (d & kg)>>6 + (d & kb)>>16
               lum / 7
               Plot(x,y,lum<<16 | lum<<8 | lum)
            Next x
         Next y
      StopDrawing()
      
      If ImgActive = #ImgWork : SCAN_FOR_PALETTE() : EndIf
      
   EndIf
   EndProcedure

Procedure CROP(image)
   Protected text$, iw, ih
      
   If selection\sFlag = #False : ProcedureReturn 0 : EndIf
   
   If IsImage(image)
      PREPARE_UNDO(image)
      
      With selection
         GrabImage(image,image,\left,\top,\right-\left+1,\bottom-\top+1)
      EndWith
      
      CANCEL_SELECTION()
      
      iw = ImageWidth(image) : ih = ImageHeight(image)
      
      If ImgActive = #ImgWork : SCAN_FOR_PALETTE() : EndIf
      
      ProcedureReturn 1
   EndIf
   
   ProcedureReturn 0
EndProcedure

Procedure SET_AS_REFERENCE()
   ; copy the working image to the reference image
   If ImgActive <> #ImgRef And IsImage(#ImgWork)
      CopyImage(#ImgRef,#imgUndoRef)
      UndoRedoStates & %1100 | #UndoRef
      CopyImage(#ImgWork,#ImgRef)
      TOG(#Opt_ViewRef)
   EndIf
EndProcedure

Procedure CLOSE_PWIN()
   ; close the palette window
   If IsWindow(#WinPalette)
      FreeImage(#imgPalette)
      CloseWindow(#WinPalette)
   EndIf
EndProcedure

Procedure CONFIG_Q_LIMIT()
   ; the quantization mode has been changed.
   ; match the 'limit' spin gadget setting to the selected mode.
   Protected Q_mode = GetGadgetState(#Combo_Qmode)
   Protected limit = GetGadgetItemData(#Combo_Qmode,Q_mode)
   
   Select Q_mode
      Case #Q_popular, #Q_adaptive
         DisableGadget(#Q_limit,#False)
         SetGadgetAttribute(#Q_limit,#PB_Spin_Maximum,256)
         SetGadgetState(#Q_limit,limit) : SetGadgetText(#Q_limit,Str(limit))
      Case #Q_uniform
         DisableGadget(#Q_limit,#False)
         SetGadgetState(#Q_limit,limit) : SetGadgetText(#Q_limit,Str(Pow(limit,3)))
         SetGadgetAttribute(#Q_limit,#PB_Spin_Maximum,16)
      Default
         Select Q_mode
            Case #Q_Fixed_16     : limit = 16
            Case #Q_fixed_24     : limit = 24
            Case #Q_fixed_32     : limit = 32
            Case #Q_neuquqnt_256 : limit = 256
            Case #BMSK_berry_128 : limit = 128
            Case #BMSK_lemon_128 : limit = 128
            Case #BMSK_ocean_128 : limit = 128
            Case #BMSK_Flat_4096 : limit = 4096
            Case #BMSK_Flat_512  : limit = 512
            Case #BMSK_flat_64   : limit = 64
            Case #BMSK_Flat_8    : limit = 8
         EndSelect
         SetGadgetText(#Q_limit,Str(limit))
         DisableGadget(#Q_limit,#True)
   EndSelect
EndProcedure

Procedure Q_LIMIT_ADJUST()
   ; update the 'limit' spin gadget.
   Protected mode = GetGadgetState(#Combo_Qmode)
   Protected limit, value, evType = EventType()
   
   If evType = #PB_EventType_Up Or evType = #PB_EventType_Down
      limit = GetGadgetState(#Q_limit)
      Select mode
         Case #Q_uniform
            SetGadgetText(#Q_limit,Str(Pow(limit,3)))
         Default
            SetGadgetText(#Q_limit,Str(limit))
      EndSelect
      SetGadgetItemData(#Combo_Qmode,mode,limit)
   ElseIf evType = #PB_EventType_Change
      If mode = #Q_uniform
         limit = GetGadgetState(#Q_limit)
         SetGadgetText(#Q_limit,Str(Pow(limit,3)))
      Else
         value = Val(GetGadgetText(#Q_limit))
         If value > 256 : SetGadgetText(#Q_limit,"256") : EndIf
         If value < 1   : SetGadgetText(#Q_limit,"2")   : EndIf
         SetGadgetState(#Q_limit,value)
         limit = GetGadgetState(#Q_limit)
         SetGadgetItemData(#Combo_Qmode,mode,limit)
      EndIf
   EndIf
   
EndProcedure

Procedure Q_LIMIT_ENTER_KEY()
   If GetActiveGadget() = #Q_limit
      PostEvent(#PB_Event_Gadget,#WinMain,#Btn_Apply,#PB_EventType_LeftClick)
   EndIf
EndProcedure

Procedure WINDOW_MOVED()
   ; reposition the palette window if the main window is moved
   Static lx,ly
   Protected x,y
   
   If IsWindow(#WinPalette)
      x = WindowX(#WinMain,#PB_Window_InnerCoordinate) + 4
      y = WindowY(#WinMain,#PB_Window_InnerCoordinate) + GadgetY(#Opt_ViewWrk) + 40
      If x <> lx Or y <> ly
         lx = x : ly = y
         ResizeWindow(#WinPalette,x,y,#PB_Ignore,#PB_Ignore)
      EndIf
   EndIf
   
EndProcedure

Procedure EXECUTE_QUANTIZATION()
   ; the quantization 'apply' button has been pressed.
   ; invoke the proper quantization method per combo box selection.
   ; the source image is #ImgRef and the output image is #ImgWork
   
   Static firstRun = #True
   Protected Q_style = GetGadgetState(#Combo_Qmode)
   
   If GetGadgetState(#Combo_Qmode) = #Q_popular Or
      GetGadgetState(#Combo_Qmode) = #Q_adaptive
      SetGadgetText(#Q_limit,Str(GetGadgetState(#Q_limit)))
   EndIf
   
   If IsImage(#ImgRef)
      If firstRun = #True
         firstRun = #False
      Else
         PREPARE_UNDO(#ImgWork)
      EndIf
      
      CopyImage(#ImgRef,#ImgWork)
   Else
      ProcedureReturn 0
   EndIf
   
   Select Q_style
      Case #Q_popular      : POPULARITY_PALETTE()
      Case #Q_uniform      : UNIFORM_PALATTE()
      Case #Q_adaptive     : ADAPTIVE_PALETTE()
      Case #BMSK_Flat_8    : BIT_MASK(Q_style)
      Case #BMSK_flat_64   : BIT_MASK(Q_style)
      Case #BMSK_Flat_512  : BIT_MASK(Q_style)
      Case #BMSK_Flat_4096 : BIT_MASK(Q_style)
      Case #BMSK_lemon_128 : BIT_MASK(Q_style)
      Case #BMSK_berry_128 : BIT_MASK(Q_style)
      Case #BMSK_ocean_128 : BIT_MASK(Q_style)
      Case #Q_Fixed_16     : FIXED_PALETTE(Q_style)
      Case #Q_fixed_24     : FIXED_PALETTE(Q_style)
      Case #Q_fixed_32     : FIXED_PALETTE(Q_style)
      Case #Q_neuquqnt_256 : NEUQUANT_256()
   EndSelect
   
   TOG(#Opt_ViewWrk); enable the 'work' option button & update canvas image
   
EndProcedure

Procedure POPULARITY_PALETTE()
   ; Create a color palette with a modified popularity approach.
   ; #ImgRef (the source) and #ImgWork (the result) must be previously defined.
   ; Required support procedures are:
   ;   'COUNT_COLORS', and 'ASSEMBLE_TO_PALETTE'.
   ; Finished palette is passed to the 'ASSEMBLE_TO_PALETTE' procedure.
   ; The palette count can be specified from 2 to 256. (256 is arbitrary limit)
   ; Theoretical maximum = 4096
   ; The higher counts will be slower.
   ; This algorithm was created by BasicallyPure.
   
   Static kb = $FF0000, kg = $00FF00, kr = $0000FF
   Protected limit = GetGadgetState(#Q_limit) ; palette will not exceed this many colors
   Protected dither = GetGadgetState(#Chk_Dither)
   Protected count, Xmax, Ymax, i, x, y, lum, d, br, da, mb, md
   
   Macro BMSK(mask) ; simple bitmask color reduction method
      StartDrawing(ImageOutput(#ImgWork))
         For y = 0 To yMax
            For x = 0 To xMax
               Plot(x, y, Point(x,y) & mask | $0F0F0F)
            Next x
         Next y
      StopDrawing()
   EndMacro
   
   If IsImage(#ImgRef) And IsImage(#ImgWork)
      count = COUNT_COLORS(#ImgWork)
      If count <= limit : limit = count : EndIf
   Else
      ProcedureReturn 0
   EndIf
   
   ; 1) preprocess colors using BitMask method, color count will
   ;    be greatly reduced.
      Xmax = ImageWidth(#ImgWork) - 1
      Ymax = ImageHeight(#ImgWork) - 1
      
      If limit < 48
         BMSK($C0C0C0) ; _64 possible
         count = COUNT_COLORS(#ImgWork)
      Else
         count = 0
      EndIf
      
      If count < (limit + 10) ; need more colors to work with
         CopyImage(#ImgRef,#ImgWork)
         BMSK($E0E0E0) ; _512 possible
         count = COUNT_COLORS(#ImgWork)
         If count < (limit + 10) ; need still more colors
            CopyImage(#ImgRef,#ImgWork)
            BMSK($F0F0F0) ; _4096 possible
         EndIf
      EndIf
   
   ; 2) gather popularity data
      NewMap Pmap.i()
      StartDrawing(ImageOutput(#ImgWork))
         Xmax = OutputWidth() - 1
         Ymax = OutputHeight() - 1
         For y = 0 To Ymax
            For x = 0 To Xmax
               Pmap(Str(Point(x,y))) + 1
            Next 
         Next
      StopDrawing()
      
      If MapSize(Pmap()) < limit : limit = MapSize(Pmap()) : EndIf
      
   ; 3) subdivide colors into 4 brightness lists
      NewList bright.PopType()
      NewList MedBri.PopType()
      NewList MedDrk.PopType()
      NewList dark.PopType()
      
      ForEach Pmap()
         d = Val(MapKey(Pmap()))
         lum = (d & kr)<<1 + (d & kg) >> 6 + (d & kb) >> 16
         If lum > 1338
            AddElement(bright()) : bright()\clr = d : bright()\pop = Pmap()
         ElseIf lum > 892
            AddElement(MedBri()) : MedBri()\clr = d : MedBri()\pop = Pmap()
         ElseIf lum > 446
            AddElement(MedDrk()) : MedDrk()\clr = d : MedDrk()\pop = Pmap()
         Else
            AddElement(dark())   : dark()\clr = d   : dark()\pop = Pmap()
         EndIf
      Next
      
   ; 4) sort each brightness lists by popularity
      SortStructuredList(bright(),#PB_Sort_Descending,OffsetOf(PopType\pop),#PB_Integer)
      SortStructuredList(MedBri(),#PB_Sort_Descending,OffsetOf(PopType\pop),#PB_Integer)
      SortStructuredList(MedDrk(),#PB_Sort_Descending,OffsetOf(PopType\pop),#PB_Integer)
      SortStructuredList(dark()  ,#PB_Sort_Descending,OffsetOf(PopType\pop),#PB_Integer)
      
   ; 5) create the final palette
      FirstElement(bright()) : br = ListSize(bright())
      FirstElement(MedBri()) : mb = ListSize(MedBri())
      FirstElement(MedDrk()) : md = ListSize(MedDrk())
      FirstElement(dark())   : da = ListSize(dark())
      
      limit - 1
      ReDim WK_Pal(limit)
      
      i = 0 : d = %00
      Repeat ; pick from each list in turn the most popular color
         If d = %00 And br > 0
            WK_Pal(i) = bright()\clr
            NextElement(bright())
            i + 1 : br - 1
         ElseIf d = %01 And da > 0
            WK_Pal(i) = dark()\clr
            NextElement(dark())
            i + 1 : da - 1
         ElseIf d = %10 And mb > 0
            WK_Pal(i) = MedBri()\clr
            NextElement(MedBri())
            i + 1 : mb - 1
         ElseIf d = %11 And md > 0
            WK_Pal(i) = MedDrk()\clr
            NextElement(MedDrk())
            i + 1 : md - 1
         EndIf
            
         d = (d + %01) & %11
      Until i > limit
      
   ; 6) create the final image
      ASSEMBLE_TO_PALETTE(WK_Pal())
   
EndProcedure

Procedure UNIFORM_PALATTE()
   ; quantize an image using evenly spaced bins
   ; the number of bins are always the cube of integer n
   ; where (n >= 2) and (n <= 16); upper limit '16' is arbitrary
   ; the number of bins sets the upper limit of possible palatte colors
   ; support procedure MN() is required
   
   Static m = #PB_Gadget_ContainerCoordinate
   Protected c,x,y,i,QL,xMax,yMax,div1,div2,div3,mlt1,mlt2,mlt3
   
   Macro Quantize(color)
      (color & $0000FF / div1 * mlt1) |
      (color & $00FF00 / div2 * mlt2) |
      (color & $FF0000 / div3 * mlt3)
   EndMacro
   
   xMax = ImageWidth(#ImgWork) - 1
   yMax = ImageHeight(#ImgWork) - 1
   
   QL = GetGadgetState(#Q_limit) ; sets the number of bins
   mlt1 = MN(QL) & $00FF ; retreive the magic multiplier
   mlt2 = mlt1 << 8
   mlt3 = mlt2 << 8
   div2 = MN(QL) & $FF00 ; retreive the magic divisor
   div1 = div2 >> 8
   div3 = div2 << 8
   
   NewMap Pmap.i() ; gathers palatte information for display only
   
   StartDrawing(ImageOutput(#ImgWork))
      DrawingMode(#PB_2DDrawing_Default)
      
      If GetGadgetState(#Chk_Dither)
         For y = 0 To yMax
            For x = 0 To xMax
               c = Quantize(PointOrdDith(x,y))
               Plot(x,y,c)
               Pmap(Str(c)) = c
            Next x
         Next y
      Else
         For y = 0 To yMax
            For x = 0 To xMax
               c = Quantize(Point(x,y))
               Plot(x,y,c)
               Pmap(Str(c)) = c
            Next x
         Next y
      EndIf
      
   StopDrawing()
   
   ; assemble the palatte and display it
   ReDim WK_Pal(MapSize(Pmap())-1)
   i = 0
   ForEach Pmap() : WK_Pal(i)=Pmap() : i+1 : Next
   
   SHOW_PALETTE(WK_Pal())

EndProcedure

Procedure ADAPTIVE_PALETTE()
   ; based on netmaestro's Adaptive Color Table procedure @
   ; http://www.purebasic.fr/english/viewtopic.php?f=12&t=35627
   ; modified for use here by BasicallyPure
   
   Protected i, x, y, cc, lastcolor, yMax, xMax
   Protected psize = GetGadgetState(#Q_limit)
   
   ReDim WK_Pal(psize-1)
   
   StartDrawing(ImageOutput(#ImgRef))
      Dim ColorBits.l(OutputWidth() * OutputHeight())
      
      xMax = OutputWidth() : yMax = OutputHeight()
      
      y = 0 : While y < yMax
         x = 0 : While x < xMax
            ColorBits(i) = Point(x,y)
            i+1
         x + 1 : Wend
      y + 1 : Wend
   StopDrawing()

   SortArray(ColorBits(),#PB_Sort_Ascending)
   x = ArraySize(colorbits())/pSize
   cc=0 
   lastcolor = colorbits(0)-1
   For i = 0 To pSize - 1
      If colorbits(cc)<>lastcolor
         WK_Pal(i) = colorbits(cc)
         lastcolor = colorbits(cc)
         cc+x
      Else
         While colorbits(cc) = lastcolor And cc < ArraySize(colorbits())
            cc+1
         Wend
         x = (ArraySize(colorbits())-cc)/(pSize-i)
         cc+x-1
         WK_Pal(i) = colorbits(cc)
         lastcolor = colorbits(cc)
      EndIf
   Next i
   
   ReDim Colorbits.l(0)
   
   ASSEMBLE_TO_PALETTE(WK_Pal())
   
EndProcedure 

Procedure FIXED_PALETTE(palette)
   ; quantize an image using one of three fixed paletts
   
   Static firstRun = #True
   Static Dim palette_16(15)
   Static Dim palette_24(23)
   Static Dim palette_32(31)
   Protected i
   
   If firstRun
      firstRun = #False
      
      Restore palData : For i = 0 To 15 : Read.i palette_16(i) : Next
      Restore palData : For i = 0 To 23 : Read.i palette_24(i) : Next
      Restore palData : For i = 0 To 31 : Read.i palette_32(i) : Next
   EndIf
   
   Select palette
      Case #Q_Fixed_16 : CopyArray(palette_16(),WK_Pal())
      Case #Q_fixed_24 : CopyArray(palette_24(),WK_Pal())
      Case #Q_fixed_32 : CopyArray(palette_32(),WK_Pal())
   EndSelect
   
   ASSEMBLE_TO_PALETTE(WK_Pal())
         
   DataSection
      palData: ; 16 colors
      ;      black    maroon   green    olive    navy     purple   teal     silver
      Data.i $000000, $000080, $008000, $008080, $800000, $800080, $808000, $C0C0C0
      ;      gray     red      lime     yellow   blue     fuchsia  aqua     white
      Data.i $808080, $0000FF, $00FF00, $00FFFF, $FF0000, $FF00FF, $FFFF00, $FFFFFF
      
      ; add 8 more for 24 colors
      ;      liteRed  sky      orange   grass    leaves   darkGreen tan      flesh_1
      Data.i $8000FF, $FFFF80, $0080FF, $40C080, $60A080, $36533C,  $3458A3, $8FAFEF
      
      ; add 8 more for 32 colors
      ;      lavender ocean    peach    pink     lemon    violet   flesh_2  flesh_3
      Data.i $FF0080, $FF8000, $8080FF, $FF80FF, $80FFFF, $FF8080, $6F8FCF, $3A4A7A
   EndDataSection
EndProcedure

Procedure NEUQUANT_256()
   ; reduce the maximum number of color to 256 using
   ; the NeuQuant Neural-Net Quantization Algorithm
   ; include file "NeuQuant_Module.pbi" is required.
   ; http://www.purebasic.fr/english/viewtopic.php?f=40&t=57939
   ; thanks to luis and wilbert for their work on the NeuQuant code
   
   Protected x, y, max_x, max_y, i
   Protected Dim Palette.l(255)
   
   If IsImage(#ImgRef)
      CopyImage(#ImgRef,#ImgWork)
      
      InitNet()
      Learn(#ImgWork)
      UnbiasNet(@Palette())
      InxBuild()
      
      StartDrawing(ImageOutput(#ImgWork))
         max_x = OutputWidth() - 1
         max_y = OutputHeight() - 1
         
         If GetGadgetState(#Chk_Dither)
            For y = 0 To max_y
               For x = 0 To max_x
                  i = InxSearch(PointOrdDith(x, y)); dither
                  Plot(x, y, Palette(i))
               Next
            Next
         Else
            For y = 0 To max_y
               For x = 0 To max_x
                  i = InxSearch(Point(x, y)); no dither
                  Plot(x, y, Palette(i))
               Next
            Next
         EndIf
         
      StopDrawing()
      
      ImgActive = #ImgWork
      SCAN_FOR_PALETTE()
   EndIf
EndProcedure

Procedure BIT_MASK(option)
   ; quantize image using bitmask method
   Protected i,x,y,p,xMax,yMax, mask
   
   StartDrawing(ImageOutput(#ImgWork))
      DrawingMode(#PB_2DDrawing_Default)
      xMax = OutputWidth() - 1
      yMax = OutputHeight() - 1
      
      Select option
         Case #BMSK_Flat_8    : mask = $808080
         Case #BMSK_flat_64   : mask = $C0C0C0
         Case #BMSK_Flat_512  : mask = $E0E0E0
         Case #BMSK_Flat_4096 : mask = $F0F0F0
         Case #BMSK_lemon_128 : mask = $80E0E0
         Case #BMSK_berry_128 : mask = $E080E0
         Case #BMSK_ocean_128 : mask = $E0E080
         Default : ProcedureReturn 0
      EndSelect
      
      NewMap Pmap.i()
      
      If GetGadgetState(#Chk_Dither)
         For y = 0 To yMax
            For x = 0 To xMax
               p = PointOrdDith(x,y) & mask | $0F0F0F
               Pmap(Str(p)) = p
               Plot(x,y,p)
            Next x
         Next y
      Else
         For y = 0 To yMax
            For x = 0 To xMax
               p = Point(x,y) & mask | $0F0F0F
               Pmap(Str(p)) = p
               Plot(x,y,p)
            Next x
         Next y
      EndIf
   StopDrawing()
   
   ReDim WK_Pal(MapSize(Pmap())-1)
   i = 0
   ForEach Pmap() : WK_Pal(i)=Pmap() : i+1 : Next
   
   SHOW_PALETTE(WK_Pal())
EndProcedure

Procedure ASSEMBLE_TO_PALETTE(Array palette(1))
   ; assign each pixel of an image to the defined palette
   ; this is slow but the only way I know how
   
   Protected Xmax, yMax, dist, near, c, d, i, x, y, R, G, B, pr, pg, pb
   Protected bins = ArraySize(palette())
   Protected dither = GetGadgetState(#Chk_Dither)
   
   CopyImage(#ImgRef,#ImgWork) ; source and output images
   
   Protected Dim pR(bins)
   Protected Dim pG(bins)
   Protected Dim pB(bins)
   
   For i = 0 To bins
      pR(i) = ((palette(i) & $0000FF)      )
      pG(i) = ((palette(i) & $00FF00) >> 8 )
      pB(i) = ((palette(i) & $FF0000) >> 16)
   Next
   
   StartDrawing(ImageOutput(#ImgWork))
      Xmax = OutputWidth() - 1 : Ymax = OutputHeight() - 1
      
      For y = 0 To Ymax
         For x = 0 To Xmax
            If dither
               c = PointOrdDith(x,y)
            Else
               c = Point(x,y)
            EndIf
            d = $FFFFFF
            R = c & $FF : c >> 8
            G = c & $FF : c >> 8
            B = c & $FF
            
            For i = 0 To bins
               dist = Abs(R-pR(i)) + Abs(G-pG(i)) + Abs(B-pB(i))
               If dist < d
                  d = dist
                  near = i
               EndIf
            Next i
            
            Plot(x,y,palette(near))
         Next x
      Next y
      
   StopDrawing()
   
   SHOW_PALETTE(palette())
EndProcedure

Procedure SET_SELECTION_BOUNDARIES(canvas)
   ; set up selection area boundaries using canvas gadget
   With selection
      \limLeft = 0
      \limTop  = 0
      \limRight  = GadgetWidth(canvas)
      \limBottom = GadgetHeight(canvas)
   EndWith
EndProcedure

Procedure SELECT_AREA(canvas)
   ; Use the mouse to select an area on the canvas gadget.
   
   Static.i x, y, lenX, lenY, lastX, lastY, boxSet
   Static.i WIC = #PB_Window_InnerCoordinate
   Protected.i drag, stretch, msx, msy, ox, oy, c, msgWin
   Protected.i Win = GetActiveWindow()
   
   With selection
      x = WindowMouseX(Win) - GadgetX(canvas,#PB_Gadget_WindowCoordinate)
      y = WindowMouseY(Win) - GadgetY(canvas,#PB_Gadget_WindowCoordinate)
      
      If x < \limLeft Or x > \limRight Or y < \limTop Or y > \limBottom
         ProcedureReturn
      EndIf
      
      If ControlKey_Down ; copy the color value under the mouse
         StartDrawing(CanvasOutput(canvas))
            c = Point(x,y)
            SetClipboardText(Str(c))
         StopDrawing()
         
         While WaitWindowEvent() <> #CanvasMouseUP : Wend
         ControlKey_Down = #False
         
         msgWin = OpenWindow(#PB_Any,0,0,200,30,"",#PB_Window_BorderLess|#PB_Window_ScreenCentered)
         TextGadget(#PB_Any,0,5,200,25,"Color value copied to clipboard.",#PB_Text_Center)
         AddWindowTimer(msgWin,0,2500)
         While WaitWindowEvent() <> #PB_Event_Timer : Wend
         CloseWindow(msgWin)
         
         ProcedureReturn
      EndIf
      
      If boxSet = #True And x > \right-9 And x < \right And y > \bottom-9 And y < \bottom
         stretch = #True ; resize selection rectangle is active
         x = lastX
         y = lastY
      ElseIf x > \left And x < \right And y > \top And y < \bottom
         drag = #True ; drage selection rectangle is active
         CompilerIf #PB_Compiler_OS = #PB_OS_Windows
            ; move cursor to center of selected area
            ox = \left + Abs(lenX / 2) + WindowX(Win, WIC) + GadgetX(#ScrollArea)
            oy = \top  + Abs(lenY / 2) + WindowY(Win, WIC) + GadgetY(#ScrollArea)
            SetCursorPos_(ox-GetGadgetAttribute(#ScrollArea,#PB_ScrollArea_X),
                          oy-GetGadgetAttribute(#ScrollArea,#PB_ScrollArea_Y))
         CompilerEndIf
      Else
         drag = #False
      EndIf
      
      \sFlag = #True
      
      While WaitWindowEvent(100) <> #CanvasMouseUP
         StartDrawing(CanvasOutput(canvas))
            DrawingMode(#PB_2DDrawing_XOr)
            If \active = #True
               ; erase previous selection rectangle
               LineXY(lastX, lastY, lastX + lenX, lastY, $FFFFFF)
               LineXY(lastX, lastY, lastX, lastY + lenY, $FFFFFF)
               LineXY(lastX, lastY + lenY, lastX + lenX, lastY + lenY, $FFFFFF)
               LineXY(lastX + lenX, lastY, lastX + lenX, lastY + lenY, $FFFFFF)
               If boxSet = #True
                  boxSet = #False
                  Box(\right - 8, \bottom - 8, 8, 8, $FFFFFF)
               EndIf
            Else
               \active = #True
               lastX = x
               lastY = y
               boxSet = #False
            EndIf
            
            msx = WindowMouseX(Win) - GadgetX(canvas,#PB_Gadget_WindowCoordinate)
            msy = WindowMouseY(Win) - GadgetY(canvas,#PB_Gadget_WindowCoordinate)
            
            ; check if mouse cursor is in bounds
            If msx >= \limLeft And msx < \limRight And msy >= \limTop And msy < \limBottom
               If drag = #False ; calculate size of next selection outline
                  lenX = msx - x
                  lenY = msy - y
               Else ; drag the selection outline
                  x = msx - lenX >> 1
                  y = msy - lenY >> 1
                  
                  ; prevent out of bounds dragging
                  If x < \limLeft : x = \limLeft : EndIf ; check left side
                  If y < \limTop  : y = \limTop  : EndIf ; check top side
                  If x > \limRight -1 - lenX ; check right side
                     x = \limRight -1 - lenX
                  EndIf
                  If y > \limBottom -1 - lenY ; check bottom side
                     y = \limBottom -1 - lenY
                  EndIf
               EndIf
            EndIf
            ; draw the selection rectangle
            LineXY(x, y, x + lenX, y, $FFFFFF)
            LineXY(x, y, x, y + lenY, $FFFFFF)
            LineXY(x, y + lenY, x + lenX, y + lenY, $FFFFFF)
            LineXY(x + lenX, y, x + lenX, y + lenY, $FFFFFF)
            
            lastX = x
            lastY = y
         StopDrawing()
         
         \left = x
         \top = y
         \right = x + lenX
         \bottom = y + lenY
         
         UPDATE_STATUS()
      Wend
      
      If \left > \right ; fix left/right crossover
         Swap \left, \right : x + lenX : lastX = x : lenX * -1
      EndIf
      
      If \top > \bottom ; fix top/bottom crossover
         Swap \top, \bottom : y + lenY : lastY = y : lenY * -1
      EndIf
      
      ; mouse button was released
      If \left = \right And \top = \bottom ; single point so no selection
         \sFlag = #False
         If ImgActive = #ImgWork
            SHOW_PALETTE(WK_Pal())
         EndIf
      Else ; a valid selection has been made
         ; draw the selection resize box
         StartDrawing(CanvasOutput(canvas))
            DrawingMode(#PB_2DDrawing_XOr)
            Box(\right - 8, \bottom - 8, 8, 8, $FFFFFF)
            boxSet = #True
         StopDrawing()
         
         If ImgActive = #ImgWork
            SCAN_FOR_PALETTE()
         EndIf
      EndIf
      
      If GetGadgetState(#Chk_Hist) : CHOOSE_HISTOGRAM() : EndIf
      
   EndWith
   
   UPDATE_STATUS()
   
EndProcedure
; IDE Options = PureBasic 5.21 LTS (Windows - x86)
; CursorPosition = 816
; FirstLine = 71
; Folding = AACwQCg8AQB9
; EnableXP
; UseIcon = CQ.ico
; Executable = ..\..\..\PureBasicExes\ColorQuantizer.exe
; DisableDebugger